#include "NavigationCostMap.h"

#include <algorithm>
#include <cmath>
#include <limits>

#include <boost/geometry.hpp>
#include <boost/geometry/algorithms/detail/intersects/interface.hpp>

#include <Eigen/Geometry>

#include <IceUtil/Time.h>

#include <VirtualRobot/BoundingBox.h>
#include <VirtualRobot/CollisionDetection/CollisionChecker.h>
#include <VirtualRobot/CollisionDetection/CollisionModel.h>
#include <VirtualRobot/Visualization/CoinVisualization/CoinVisualizationFactory.h>
#include <VirtualRobot/XML/ObjectIO.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <armarx/navigation/conversions/eigen.h>

namespace bg = boost::geometry;

namespace armarx::navigation::algorithm::astar
{

    NavigationCostMap::NavigationCostMap(VirtualRobot::RobotPtr robot,
                                         VirtualRobot::SceneObjectSetPtr obstacles,
                                         const Parameters& parameters) :
        robot(robot), obstacles(obstacles), parameters(parameters)
    {
    }

    NavigationCostMap::Position
    NavigationCostMap::toPosition(const NavigationCostMap::Index& index) const
    {
        Eigen::Vector2f pos;
        // TODO check if x and y must be switched
        pos.x() = sceneBoundsMin.x() + index.x() * parameters.cellSize + parameters.cellSize / 2;
        pos.y() = sceneBoundsMin.y() + index.y() * parameters.cellSize + parameters.cellSize / 2;

        return pos;
    }

    void
    NavigationCostMap::createUniformGrid()
    {
        ARMARX_TRACE;

        if (obstacles)
        {
            for (size_t i = 0; i < obstacles->getCollisionModels().size(); i++)
            {
                VirtualRobot::BoundingBox bb = obstacles->getCollisionModels()[i]->getBoundingBox();
                sceneBoundsMin.x() = std::min(bb.getMin().x(), sceneBoundsMin.x());
                sceneBoundsMin.y() = std::min(bb.getMin().y(), sceneBoundsMin.y());
                sceneBoundsMax.x() = std::max(bb.getMax().x(), sceneBoundsMax.x());
                sceneBoundsMax.y() = std::max(bb.getMax().y(), sceneBoundsMax.y());
            }
        }

        ARMARX_INFO << "Scene bounds are " << sceneBoundsMin << " and " << sceneBoundsMax;

        //+1 for explicit rounding up
        c_x = (sceneBoundsMax.x() - sceneBoundsMin.x()) / parameters.cellSize + 1;
        c_y = (sceneBoundsMax.y() - sceneBoundsMin.y()) / parameters.cellSize + 1;

        ARMARX_INFO << "Grid size: " << c_y << ", " << c_x;

        ARMARX_INFO << "Resetting grid";
        grid = Eigen::MatrixXf(c_x, c_y); //.array();
        // grid = Eigen::ArrayXf(); //(c_x, c_y);
        // grid->resize(c_x, c_y);

        ARMARX_INFO << "done.";
    }


    float
    NavigationCostMap::computeCost(const NavigationCostMap::Position& position)
    {
        ARMARX_TRACE;
        ARMARX_CHECK_NOT_NULL(robotCollisionModel);
        ARMARX_CHECK_NOT_NULL(VirtualRobot::CollisionChecker::getGlobalCollisionChecker());
        ARMARX_CHECK_NOT_NULL(obstacles);

        const core::Pose globalPose = Eigen::Translation3f(conv::to3D(position)) *
                                      Eigen::AngleAxisf(M_PI_2f32, Eigen::Vector3f::UnitX());

        // auto colModel = robotCollisionModel->clone();
        // colModel->setGlobalPose(globalPose.matrix());

        robotCollisionModel->setGlobalPose(globalPose.matrix());

        // Unused.
        if (false)
        {
            VirtualRobot::BoundingBox robotBbox = robotCollisionModel->getBoundingBox(true);
            (void) robotBbox;
        }

        Eigen::Vector3f P1;
        Eigen::Vector3f P2;
        int id1, id2;

        float minDistance = std::numeric_limits<float>::max();

        // TODO omp...
        for (size_t i = 0; i < obstacles->getSize(); i++)
        {
            // cheap collision check
            // VirtualRobot::BoundingBox obstacleBbox =
            //     obstacles->getSceneObject(i)->getCollisionModel()->getBoundingBox(true);
            // if (not intersects(robotBbox, obstacleBbox))
            // {
            //     continue;
            // }

            // precise collision check
            const float dist =
                VirtualRobot::CollisionChecker::getGlobalCollisionChecker()->calculateDistance(
                    robotCollisionModel,
                    obstacles->getSceneObject(i)->getCollisionModel(),
                    P1,
                    P2,
                    &id1,
                    &id2);

            // check if objects collide
            if ((dist <= parameters.cellSize / 2) or
                VirtualRobot::CollisionChecker::getGlobalCollisionChecker()->checkCollision(
                    robotCollisionModel, obstacles->getSceneObject(i)->getCollisionModel()))
            {
                minDistance = 0;
                break;
            }

            minDistance = std::min(minDistance, dist);
        }
        // return n->position.x() >= sceneBoundsMin.x() && n->position.x() <= sceneBoundsMax.x() &&
        //        n->position.y() >= sceneBoundsMin.y() && n->position.y() <= sceneBoundsMax.y();

        return minDistance;
    }

    NavigationCostMap::Vertex
    NavigationCostMap::toVertex(Eigen::Vector2f v)
    {
        float v_y = (v.y() - parameters.cellSize / 2 - sceneBoundsMin.y()) / parameters.cellSize;
        float v_x = (v.x() - parameters.cellSize / 2 - sceneBoundsMin.x()) / parameters.cellSize;

        ARMARX_CHECK(v_y >= 0.F);
        ARMARX_CHECK(v_x >= 0.F);

        ARMARX_CHECK(v_y <= (c_y - 1));
        ARMARX_CHECK(v_x <= (c_x - 1));

        return Vertex{.index = Index{static_cast<int>(v_x), static_cast<int>(v_y)}, .position = v};
    }

    void
    NavigationCostMap::createCostmap()
    {
        ARMARX_TRACE;
        std::vector<Eigen::Vector2f> result;

        ARMARX_CHECK_NOT_NULL(robot) << "Robot must be set";


        VirtualRobot::CoinVisualizationFactory factory;
        const auto cylinder = factory.createCylinder(1500.F / 2, 2000.F);

        VirtualRobot::CollisionModelPtr collisionModel(new VirtualRobot::CollisionModel(cylinder));


        robotCollisionModel = collisionModel;


        ARMARX_CHECK_NOT_NULL(robotCollisionModel) << "Collision model must be set!";

        // ARMARX_CHECK(robot->hasRobotNode(robotColModelName)) << "No robot node " << robotColModelName;
        // ARMARX_CHECK(robot->getRobotNode(robotColModelName)->getCollisionModel())
        // << "No collision model for robot node " << robotColModelName;

        // robotCollisionModel = robot->getRobotNode(robotColModelName)->getCollisionModel()->clone();
        robotCollisionModel->setGlobalPose(robot->getGlobalPose());

        // static auto cylinder = VirtualRobot::ObjectIO::loadObstacle("/home/fabi/cylinder.xml");
        // ARMARX_CHECK_NOT_NULL(cylinder);
        // cylinder->setGlobalPose(robot->getGlobalPose());

        // robotCollisionModel = cylinder->getCollisionModel();

        ARMARX_CHECK_NOT_NULL(robotCollisionModel);

        ARMARX_INFO << "Creating grid";
        createUniformGrid();

        ARMARX_INFO << "Filling grid";
        fillGridCosts();

        ARMARX_INFO << "Filled grid";
    }

    void
    NavigationCostMap::setRobotCollisionModel(const VirtualRobot::CollisionModelPtr& collisionModel)
    {
        robotCollisionModel = collisionModel;
    }

    void
    NavigationCostMap::fillGridCosts()
    {
        int i = 0;

        for (unsigned int x = 0; x < c_x; x++)
        {
            for (unsigned int y = 0; y < c_y; y++)
            {
                const Index index{x, y};
                const Position position = toPosition(index);

                grid.value()(x, y) = computeCost(position);

                i++;

                ARMARX_INFO << i << "/" << c_x * c_y;
            }
        }
    }

} // namespace armarx::navigation::algorithm::astar
